@@ -1,7 +1,8 @@
 /* dynlist.c - dynamic list overlay */
+/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/dynlist.c,v 1.5.2.9 2007/01/02 21:44:08 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2003-2006 The OpenLDAP Foundation.
+ * Copyright 2003-2007 The OpenLDAP Foundation.
  * Portions Copyright 2004-2005 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -22,11 +23,15 @@
 
 #ifdef SLAPD_OVER_DYNLIST
 
-#if LDAP_VENDOR_VERSION_MINOR != X && LDAP_VENDOR_VERSION_MINOR < 3
-#define OL_2_2_COMPAT
-#elif defined(LDAP_DEVEL) && SLAPD_OVER_DYNGROUP != SLAPD_MOD_STATIC
+#if LDAP_VENDOR_VERSION_MINOR == X || LDAP_VENDOR_VERSION_MINOR > 3
+#if SLAPD_OVER_DYNGROUP != SLAPD_MOD_STATIC
 #define TAKEOVER_DYNGROUP
 #endif
+#else
+#if LDAP_VENDOR_VERSION_MINOR < 3
+#define OL_2_2_COMPAT
+#endif
+#endif
 
 #include <stdio.h>
 
@@ -350,7 +355,7 @@ dynlist_send_entry( Operation *op, SlapReply *rs, dynlist_info_t *dli )
 			continue;
 		}
 
-		if ( lud->lud_host ) {
+		if ( lud->lud_host != NULL ) {
 			/* FIXME: host not allowed; reject as illegal? */
 			Debug( LDAP_DEBUG_ANY, "dynlist_send_entry(\"%s\"): "
 				"illegal URI \"%s\"\n",
@@ -365,6 +370,7 @@ dynlist_send_entry( Operation *op, SlapReply *rs, dynlist_info_t *dli )
 			 * this can be useful in case of a database serving
 			 * the empty suffix */
 			BER_BVSTR( &dn, "" );
+
 		} else {
 			ber_str2bv( lud->lud_dn, 0, 0, &dn );
 		}
@@ -435,6 +441,7 @@ dynlist_send_entry( Operation *op, SlapReply *rs, dynlist_info_t *dli )
 		if ( lud->lud_filter == NULL ) {
 			ber_dupbv_x( &o.ors_filterstr,
 					&dli->dli_default_filter, op->o_tmpmemctx );
+
 		} else {
 			struct berval	flt;
 			ber_str2bv( lud->lud_filter, 0, 0, &flt );
@@ -471,13 +478,9 @@ cleanup:;
 		if ( !BER_BVISNULL( &o.o_req_ndn ) ) {
 			op->o_tmpfree( o.o_req_ndn.bv_val, op->o_tmpmemctx );
 		}
-		if ( o.ors_filterstr.bv_val != lud->lud_filter ) {
-			op->o_tmpfree( o.ors_filterstr.bv_val, op->o_tmpmemctx );
-			lud->lud_filter = NULL;
-		}
-		if ( lud ) {
-			ldap_free_urldesc( lud );
-		}
+		assert( o.ors_filterstr.bv_val != lud->lud_filter );
+		op->o_tmpfree( o.ors_filterstr.bv_val, op->o_tmpmemctx );
+		ldap_free_urldesc( lud );
 	}
 
 	rs->sr_entry = e;
@@ -696,7 +699,7 @@ dynlist_build_def_filter( dynlist_info_t *dli )
 
 	dli->dli_default_filter.bv_len = STRLENOF( "(!(objectClass=" "))" )
 		+ dli->dli_oc->soc_cname.bv_len;
-	dli->dli_default_filter.bv_val = SLAP_MALLOC( dli->dli_default_filter.bv_len + 1 );
+	dli->dli_default_filter.bv_val = ch_malloc( dli->dli_default_filter.bv_len + 1 );
 	if ( dli->dli_default_filter.bv_val == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "dynlist_db_open: malloc failed.\n",
 			0, 0, 0 );
@@ -1345,9 +1348,11 @@ dynlist_db_open(
 			dli->dli_ad = ad;			
 		}
 
-		rc = dynlist_build_def_filter( dli );
-		if ( rc != 0 ) {
-			return rc;
+		if ( BER_BVISNULL( &dli->dli_default_filter ) ) {
+			rc = dynlist_build_def_filter( dli );
+			if ( rc != 0 ) {
+				return rc;
+			}
 		}
 	}
 
