@@ -1,8 +1,8 @@
 /* rwmmap.c - rewrite/mapping routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/rwmmap.c,v 1.14.2.10 2006/01/03 22:16:25 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/rwmmap.c,v 1.14.2.12 2007/01/03 08:55:04 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2006 The OpenLDAP Foundation.
+ * Copyright 1999-2007 The OpenLDAP Foundation.
  * Portions Copyright 1999-2003 Howard Chu.
  * Portions Copyright 2000-2003 Pierangelo Masarati.
  * All rights reserved.
@@ -449,6 +449,7 @@ rwm_int_filter_map_rewrite(
 {
 	int		i;
 	Filter		*p;
+	AttributeDescription *ad;
 	struct berval	atmp,
 			vtmp,
 			*tmp;
@@ -475,7 +476,8 @@ rwm_int_filter_map_rewrite(
 
 	switch ( f->f_choice ) {
 	case LDAP_FILTER_EQUALITY:
-		if ( map_attr_value( dc, &f->f_av_desc, &atmp,
+		ad = f->f_av_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					&f->f_av_value, &vtmp, RWM_MAP ) )
 		{
 			goto computed;
@@ -491,7 +493,8 @@ rwm_int_filter_map_rewrite(
 		break;
 
 	case LDAP_FILTER_GE:
-		if ( map_attr_value( dc, &f->f_av_desc, &atmp,
+		ad = f->f_av_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					&f->f_av_value, &vtmp, RWM_MAP ) )
 		{
 			goto computed;
@@ -507,7 +510,8 @@ rwm_int_filter_map_rewrite(
 		break;
 
 	case LDAP_FILTER_LE:
-		if ( map_attr_value( dc, &f->f_av_desc, &atmp,
+		ad = f->f_av_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					&f->f_av_value, &vtmp, RWM_MAP ) )
 		{
 			goto computed;
@@ -523,7 +527,8 @@ rwm_int_filter_map_rewrite(
 		break;
 
 	case LDAP_FILTER_APPROX:
-		if ( map_attr_value( dc, &f->f_av_desc, &atmp,
+		ad = f->f_av_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					&f->f_av_value, &vtmp, RWM_MAP ) )
 		{
 			goto computed;
@@ -539,7 +544,8 @@ rwm_int_filter_map_rewrite(
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
-		if ( map_attr_value( dc, &f->f_sub_desc, &atmp,
+		ad = f->f_sub_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					NULL, NULL, RWM_MAP ) )
 		{
 			goto computed;
@@ -601,7 +607,8 @@ rwm_int_filter_map_rewrite(
 		break;
 
 	case LDAP_FILTER_PRESENT:
-		if ( map_attr_value( dc, &f->f_desc, &atmp,
+		ad = f->f_desc;
+		if ( map_attr_value( dc, &ad, &atmp,
 					NULL, NULL, RWM_MAP ) )
 		{
 			goto computed;
@@ -647,7 +654,8 @@ rwm_int_filter_map_rewrite(
 
 	case LDAP_FILTER_EXT: {
 		if ( f->f_mr_desc ) {
-			if ( map_attr_value( dc, &f->f_mr_desc, &atmp,
+			ad = f->f_mr_desc;
+			if ( map_attr_value( dc, &ad, &atmp,
 						&f->f_mr_value, &vtmp, RWM_MAP ) )
 			{
 				goto computed;
